# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# ruff: noqa: ERA001

project = "Rapporto"
copyright = "2025, The Panodata Developers"  # noqa: A001
author = "The Panodata Developers"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.ifconfig",
    "sphinxcontrib.mermaid",
    "sphinxext.opengraph",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_title = "Rapporto"

# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/announcements.html
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/layout.html#build-date
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/header-links.html
# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/source-buttons.html
html_theme_options = {
    "announcement": """
    Rapporto is in its infancy, and needs your support. <br>
    We appreciate any kind of
    <a href="https://github.com/tech-writing/rapporto/issues">bug report</a> and
    <a href="https://github.com/tech-writing/rapporto/issues">feature request</a>,
    or suggestions to improve its
    <a href="https://github.com/tech-writing/rapporto/tree/main/docs">documentation</a>.
    """,
    "content_footer_items": ["last-updated"],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/tech-writing/rapporto",
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/rapporto",
            "icon": "fa-custom fa-pypi",
        },
    ],
    "icon_links_label": "Quick Links",
    "navbar_align": "left",
    "show_toc_level": 1,
    "use_edit_page_button": True,
}

# https://pydata-sphinx-theme.readthedocs.io/en/stable/user_guide/source-buttons.html#github
html_context = {
    "github_user": "tech-writing",
    "github_repo": "rapporto",
    "github_version": "main",
    "doc_path": "docs",
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False
html_copy_source = False

# -- Intersphinx ----------------------------------------------------------

intersphinx_mapping = {
    # "influxio": ("https://influxio.readthedocs.io/", None),
}
linkcheck_ignore = []

# Disable caching remote inventories completely.
# http://www.sphinx-doc.org/en/stable/ext/intersphinx.html#confval-intersphinx_cache_limit
# intersphinx_cache_limit = 0

language = "en"

# -- Extension configuration -------------------------------------------------

sphinx_tabs_valid_builders = ["linkcheck"]
todo_include_todos = True

# Configure sphinx-copybutton
copybutton_remove_prompts = True
copybutton_line_continuation_character = "\\"
copybutton_prompt_text = (
    r">>> |\.\.\. |\$ |sh\$ |PS> |cr> |mysql> |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True

# Configure sphinxext-opengraph
ogp_site_url = "https://rapporto.readthedocs.io/"
ogp_enable_meta_description = True
# ogp_image = "http://example.org/image.png"
# ogp_description_length = 300


# -- Options for MyST -------------------------------------------------

myst_heading_anchors = 3
myst_enable_extensions = [
    "attrs_block",
    "attrs_inline",
    "colon_fence",
    "deflist",
    "fieldlist",
    "html_admonition",
    "html_image",
    "linkify",
    "replacements",
    "strikethrough",
    "substitution",
    "tasklist",
]
myst_substitutions = {}



import datetime as dt
import logging
import os
import typing as t
from collections import OrderedDict
from operator import attrgetter
from textwrap import dedent

import requests_cache
from munch import munchify
from tqdm import tqdm

from rapporto.github.model import (
    ActionsFilter,
    ActionsOutcome,
    ActivityInquiry,
    GitHubActivityQueryBuilder,
    GitHubAttentionQueryBuilder,
    GitHubSearch,
    MultiRepositoryInquiry,
    PullRequestMetadata,
)
from rapporto.github.util import repository_name
from rapporto.util import sanitize_title

logger = logging.getLogger(__name__)


class HttpClient:
    session = requests_cache.CachedSession(backend="sqlite", expire_after=3600)
    if "GH_TOKEN" in os.environ:
        session.headers.update({"Authorization": f"Bearer {os.getenv('GH_TOKEN')}"})
    else:
        logger.warning("GH_TOKEN not defined. This will exhaust the rate limit quickly.")


class GitHubActivityReport:
    """
    Report about activity across a whole GitHub organization.
    """

    def __init__(self, inquiry: ActivityInquiry):
        self.inquiry = inquiry
        self.session = HttpClient.session
        self.search = GitHubSearch.with_query_builder(
            self.session, GitHubActivityQueryBuilder(inquiry=inquiry)
        )

    @property
    def repository_names(self):
        items = self.search.issues_and_prs()
        names = []
        for item in items:
            names.append(repository_name(item["repository_url"]))
        return sorted(set(names))

    def pull_requests(self):
        items = []
        response = self.session.get(self.search.pulls_api)
        response.raise_for_status()
        for item in tqdm(response.json()["items"], leave=False):
            if "pull_request" in item:
                pr_url = item["pull_request"]["url"]
                response = self.session.get(pr_url)
                response.raise_for_status()
                pr_data = response.json()
                cm = PullRequestMetadata.from_dict(pr_data)  # type: ignore[attr-defined]
                items.append(cm)
        return items

    def significant_pull_requests(self):
        """
        Return 2/5 of the most significant PRs, or any other share.

        The list of candidates (all PRs within given time range) is sorted by
        number of comments, number of changed files, and delta code size.
        """
        prs = self.pull_requests()
        items_max = int(len(prs) / 2 / 5) + 1
        items = []

        by_size_sort_attributes = ["code_size", "changed_files", "comments_total"]
        by_size_items = sorted(prs, key=attrgetter(*by_size_sort_attributes), reverse=True)
        items += by_size_items[:items_max]

        by_comments_sort_attributes = ["comments_total", "changed_files", "code_size"]
        by_comments_items = sorted(prs, key=attrgetter(*by_comments_sort_attributes), reverse=True)
        # TODO: Need to deduplicate manually.
        # items += by_comments_items[:items_max]
        for item in by_comments_items[:items_max]:
            if item not in items:
                items.append(item)

        return items

    @property
    def markdown_overview(self):
        link_issues = f"[Issues]({self.search.issues_html})"
        link_pulls = f"[Pull requests]({self.search.pulls_html})"
        return dedent(f"""
        *Progress:*
          - About: Bugfixes, Documentation, Guidance, Planning, Support
          - Activity: {", ".join(self.repository_names)}
          - Details: {link_issues}, {link_pulls}
        *Plans:* Dito.
        *Problems:* n/a
        """).strip()

    @property
    def markdown_significant(self):
        items = [item.format_item() for item in self.significant_pull_requests()]
        return "\n".join(items)

    @staticmethod
    def format_pr(item: PullRequestMetadata):
        return (
            f"comments: {item.comments_total}, files: {item.changed_files}, size: {item.code_size}"
        )

    def print(self):
        timerange = self.inquiry.created and f"for {self.inquiry.created}" or ""
        print(f"# PPP report {timerange}")
        # print("## Overview")
        print(self.markdown_overview)
        print()
        print("*Top changes:*")
        print(self.markdown_significant)


class GitHubActionsReport:
    """
    Report about failed outcomes of GitHub Actions workflow runs.
    """

    DELTA_HOURS = 24

    def __init__(self, inquiry: MultiRepositoryInquiry):
        self.inquiry = inquiry
        self.session = HttpClient.session

    @property
    def yesterday(self) -> str:
        """
        Compute the start timestamp in ISO format.
        Truncate the ISO format after the hour, to permit caching.
        """
        return dt.datetime.strftime(
            dt.datetime.now() - dt.timedelta(hours=self.DELTA_HOURS), "%Y-%m-%dT%H"
        )

    def fetch(self, filter: ActionsFilter) -> t.List[ActionsOutcome]:  # noqa:A002
        outcomes = []
        for repository in tqdm(
            self.inquiry.repositories,
            desc=f"Fetching failed GitHub Actions outcomes for event={filter.event}",
            leave=False,
        ):
            url = f"https://api.github.com/repos/{repository}/actions/runs?{filter.query}"
            logger.debug(f"Using API URL: {url}")
            response = self.session.get(url)
            if response.status_code == 404:
                continue
            response.raise_for_status()
            runs = munchify(response.json()).workflow_runs
            for run in runs:
                outcome = ActionsOutcome(
                    status=run.status,
                    repository=repository,
                    name=run.display_title,
                    url=run.html_url,
                    started=run.run_started_at,
                )
                outcomes.append(outcome)
        return outcomes

    def to_markdown(self, filter: ActionsFilter) -> str:  # noqa:A002
        return "\n".join([item.markdown for item in self.fetch(filter=filter)])

    @property
    def markdown(self):
        items_scheduled = self.to_markdown(
            ActionsFilter(event="schedule", status="failure", created=f">{self.yesterday}")
        )
        items_pull_requests = self.to_markdown(
            ActionsFilter(event="pull_request", status="failure", created=f">{self.yesterday}")
        )
        items_dynamic = self.to_markdown(
            ActionsFilter(event="dynamic", status="failure", created=f">{self.yesterday}")
        )
        return dedent(f"""
# CI failures report {dt.datetime.now().strftime("%Y-%m-%d")}
A report about GitHub Actions workflow runs that failed recently (now-{self.DELTA_HOURS}h).

## Scheduled
{items_scheduled or "n/a"}

## Pull requests
{items_pull_requests or "n/a"}

## Dynamic
{items_dynamic or "n/a"}
        """).strip()  # noqa: E501

    def print(self):
        print(self.markdown)


class GitHubAttentionReport:
    """
    Report about important items that deserve your attention, bugs first.

    Find all issues and pull requests with labels "bug" or "important".
    """

    def __init__(self, inquiry: ActivityInquiry):
        self.inquiry = inquiry
        self.session = HttpClient.session
        self.search = GitHubSearch.with_query_builder(
            self.session, GitHubAttentionQueryBuilder(inquiry=inquiry)
        )

    @property
    def markdown(self):
        items = self.search.issues_and_prs()
        items = sorted(munchify(items), key=attrgetter("created_at"))

        bug = []
        important = []
        other = []
        for item in tqdm(items, leave=False):
            labels = [label.name for label in item.labels]
            title = sanitize_title(
                f"{repository_name(item.repository_url, with_org=True)}: {item.title}"
            )
            link = f"[{title}]({item.html_url})"
            line = f"- {link}"
            # line = f"- {link} {', '.join(labels)}"
            if "bug" in labels or "type-bug" in labels or "type-crash" in labels:
                bug.append(line)
                continue
            if "important" in labels:
                important.append(line)
                continue
            other.append(line)

        sections = OrderedDict()
        if bug:
            sections["Bugs"] = "\n".join(bug)
        if important:
            sections["Important"] = "\n".join(important)
        if other:
            sections["Others"] = "\n".join(other)

        def render_section(name) -> str:
            if name not in sections:
                return ""
            return f"\n## {name}\n{sections[name]}"

        return dedent(f"""
# Importance report {self.inquiry.created or ""}

A report about important items that deserve your attention, bugs first.
Time range: {self.search.query_builder.timerange or "n/a"}
{render_section("Bugs")}
{render_section("Important")}
{render_section("Others")}
        """).strip()  # noqa: E501

    def print(self):
        print(self.markdown)



import urllib.parse
from pathlib import Path


def repository_name(url: str, with_org: bool = False) -> str:
    uri = urllib.parse.urlparse(url)
    path = Path(uri.path)
    if with_org:
        return path.parent.name + "/" + path.name
    else:
        return path.name



import typing as t
from pathlib import Path

import click
from click_aliases import ClickAliasedGroup

from rapporto.github.core import GitHubActionsReport, GitHubActivityReport, GitHubAttentionReport
from rapporto.github.model import ActivityInquiry, MultiRepositoryInquiry

organization_option = click.option("--organization", "--org", type=str, required=False)
author_option = click.option("--author", type=str, required=False)
timerange_option = click.option("--timerange", type=str, required=False)
repository_option = click.option("--repository", type=str, required=False)
repositories_file_option = click.option("--repositories-file", type=Path, required=False)


@click.group(cls=ClickAliasedGroup)
@click.pass_context
def cli(ctx: click.Context):
    """
    Harvest information from GitHub.
    """
    pass


@cli.command()
@organization_option
@author_option
@timerange_option
def ppp(
    organization: t.Optional[str] = None,
    author: t.Optional[str] = None,
    timerange: t.Optional[str] = None,
):
    """
    Activity: Report about activities of individual authors.
    """
    inquiry = ActivityInquiry(organization=organization, author=author, created=timerange)
    report = GitHubActivityReport(inquiry=inquiry)
    report.print()


@cli.command()
@repository_option
@repositories_file_option
def ci(repository: str, repositories_file: Path = None):
    """
    QA: Report about CI failures.
    """
    try:
        inquiry = MultiRepositoryInquiry.make(
            repository=repository, repositories_file=repositories_file
        )
    except ValueError as ex:
        click.echo(
            f"Please specify valid input for --repository or --repositories-file: {ex}", err=True
        )
        raise SystemExit(1) from ex
    report = GitHubActionsReport(inquiry=inquiry)
    report.print()


@cli.command(aliases=["att"])
@organization_option
@timerange_option
def attention(organization: t.Optional[str] = None, timerange: t.Optional[str] = None):
    """
    QA: Report about important items that deserve attention.
    """
    inquiry = ActivityInquiry(organization=organization, created=timerange)
    report = GitHubAttentionReport(inquiry=inquiry)
    report.print()



from rapporto.cli import cli


def test_cli_version(cli_runner):
    """
    CLI test: Invoke `rapporto --version`.
    """
    result = cli_runner.invoke(
        cli,
        args="--version",
        catch_exceptions=False,
    )
    assert result.exit_code == 0



import logging
import os
import re
from datetime import datetime

import requests
from slack_sdk import WebClient as SlackClient
from slack_sdk.errors import SlackApiError

logger = logging.getLogger(__name__)


class SlackThreadExporter:
    """
    Export a Slack thread and convert into Markdown format.
    """

    def __init__(self, token):
        self.client = SlackClient(token)
        self.user_cache = {}

    def resolve_user_id(self, user_id):
        """
        Resolve a Slack user ID to a username.

        :param user_id: Slack user ID (e.g., 'U01S5H7RRGB')
        :return: Username string (e.g., '@john.doe') or the original mention if unresolved
        """
        if user_id in self.user_cache:
            logger.debug(f"User ID {user_id} found in cache.")
            return self.user_cache[user_id]

        try:
            user_info = self.client.users_info(user=user_id)
            if user_info["ok"]:
                username = f"@{user_info['user']['name']}"  # Using 'name' for @username format
                self.user_cache[user_id] = username
                logger.debug(f"Resolved user ID {user_id} to username {username}.")
                return username
            else:
                logger.error(f"Failed to resolve user ID {user_id}: {user_info['error']}")
                return f"<@{user_id}>"  # Return original mention if failed
        except SlackApiError as e:
            logger.error(f"Slack API error when resolving user ID {user_id}: {e.response['error']}")
            return f"<@{user_id}>"  # Return original mention if exception occurs

    def replace_user_mentions(self, text):
        """
        Replace Slack user mentions in the format <@USERID> with @username in bold.

        :param text: Original message text
        :return: Text with user mentions replaced by bold usernames
        """
        pattern = re.compile(r"<@([A-Z0-9]+)>")  # Matches patterns like <@U01S5H7RRGB>
        matches = pattern.findall(text)
        logger.debug(f"Found user mentions: {matches}")
        for user_id in matches:
            username = self.resolve_user_id(user_id)
            text = text.replace(f"<@{user_id}>", f"**{username}**")
            logger.debug(f"Replaced <@{user_id}> with **{username}**")
        return text

    def export_thread(self, slack_link, output_dir="slack_export"):
        """
        Export a Slack thread to markdown.

        :param slack_link: Slack thread URL
        :param output_dir: Directory to save exported files
        """
        logger.info(f"Starting export for Slack link: {slack_link}")

        # Parse link to get channel and thread details
        try:
            channel_id, thread_ts = self._parse_slack_link(slack_link)
            logger.debug(f"Parsed channel ID: {channel_id}, Thread Timestamp: {thread_ts}")
        except ValueError as ve:
            logger.error(f"Error parsing Slack link: {ve}")
            return

        # Fetch channel name
        try:
            channel_info = self.client.conversations_info(channel=channel_id)
            channel_name = channel_info["channel"]["name"]
            logger.debug(f"Fetched channel name: {channel_name}")
        except SlackApiError as e:
            logger.error(f"Error fetching channel info: {e.response['error']}")
            return

        # Fetch thread messages
        try:
            thread_replies = self.client.conversations_replies(channel=channel_id, ts=thread_ts)
            logger.debug(f"Fetched {len(thread_replies['messages'])} messages in thread.")
        except SlackApiError as e:
            logger.error(f"Error fetching thread: {e.response['error']}")
            return

        # Get the title from the first message
        title = thread_replies["messages"][0].get("text", "")[:20]
        logger.debug(f"Extracted title from first message: {title}")

        # Create output directories
        os.makedirs(output_dir, exist_ok=True)
        attachments_dir = os.path.join(output_dir, "attachments")
        os.makedirs(attachments_dir, exist_ok=True)
        logger.debug(f"Created output directories: {output_dir} and {attachments_dir}")

        # Generate sanitized filename
        sanitized_channel_name = self._sanitize_filename(channel_name)
        sanitized_title = self._sanitize_filename(title)
        sanitized_ts = datetime.fromtimestamp(float(thread_ts)).strftime("%y-%m-%d-%H_%M")
        filename_md = f"{sanitized_channel_name}_{sanitized_title}_{sanitized_ts}.md"
        logger.debug(f"Generated markdown filename: {filename_md}")

        # Markdown content
        content = []
        file_counter = 1

        # Process each message in the thread
        for message in thread_replies["messages"]:
            # Parse message details
            user_id = message.get("user", "")
            if user_id:
                username = self.resolve_user_id(user_id)
            else:
                username = "System"  # For messages like system messages without a user ID
                logger.debug("Message without user ID detected. Marked as 'System'.")

            timestamp = datetime.fromtimestamp(float(message["ts"])).strftime("%Y-%m-%d %H:%M:%S")
            content.append(f"### {username} - {timestamp}")
            logger.debug(f"Added header for message by {username} at {timestamp}.")

            # Process message text with formatting
            text = message.get("text", "")
            text = text.replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")
            text = text.replace("```", "\n```")
            logger.debug(f"Processed message text: {text}")

            # Replace user mentions with usernames (formatted in bold)
            text = self.replace_user_mentions(text)
            logger.debug(f"Replaced user mentions in text: {text}")

            # Check if the message is from Opsgenie and format accordingly
            logger.debug(f"Checking if message is from Opsgenie: {username}")
            if "opsgenie" in username.lower():
                logger.debug("Identified as Opsgenie message.")
                # Check if the message has attachments or blocks
                if "attachments" in message:
                    for attachment in message["attachments"]:
                        if "text" in attachment:
                            text += "\n" + attachment["text"]
                            logger.debug("Appended attachment text from Opsgenie message.")
                if "blocks" in message:
                    for block in message["blocks"]:
                        if "text" in block and "text" in block["text"]:
                            text += "\n" + block["text"]["text"]
                            logger.debug("Appended block text from Opsgenie message.")
                text = self._format_opsgenie_message(text)
                logger.debug(f"Formatted Opsgenie message text: {text}")

            content.append(text)

            # Handle file attachments
            if "files" in message:
                for file_info in message["files"]:
                    original_filename = file_info.get("name", "unnamed")
                    sanitized_filename = self._sanitize_filename(original_filename)
                    basename, extension = os.path.splitext(sanitized_filename)

                    # Create unique filename with sanitized_ts and counter
                    unique_filename = f"{basename}_{sanitized_ts}_{file_counter}{extension}"
                    file_path = os.path.join(attachments_dir, unique_filename)
                    logger.debug(f"Prepared to download file: {unique_filename}")

                    # Download file with authorization
                    headers = {"Authorization": f"Bearer {self.client.token}"}
                    success = self._download_file(
                        file_info.get("url_private_download", ""), file_path, headers
                    )

                    if success:
                        # Add file reference to markdown
                        if file_info.get("mimetype", "").startswith("image/"):
                            content.append(f"![{unique_filename}](attachments/{unique_filename})")
                            logger.debug(f"Embedded image in markdown: {unique_filename}")
                        else:
                            attachment_link = (
                                f"[Download {unique_filename}](attachments/{unique_filename})"
                            )
                            content.append(f"**Attachment:** {attachment_link}")
                            logger.debug(f"Added attachment link in markdown: {unique_filename}")
                        file_counter += 1  # Increment counter after successful download
                    else:
                        logger.error(f"Failed to download file: {unique_filename}")

            # Handle reactions
            if "reactions" in message:
                reactions_list = []
                for reaction in message["reactions"]:
                    emoji = reaction["name"]
                    users = reaction["users"]
                    # Resolve usernames from user IDs
                    usernames = [self.resolve_user_id(user_id) for user_id in users]
                    # Join usernames with commas
                    usernames_str = ", ".join(usernames)
                    # Add to reactions list
                    reactions_list.append(f":{emoji}: {usernames_str}")
                    logger.debug(f"Processed reaction: :{emoji}: by {usernames_str}")
                # Append reactions section to markdown_content
                if reactions_list:
                    reactions_md = "\n**Reactions:**\n"
                    for reaction_entry in reactions_list:
                        reactions_md += f"- {reaction_entry}\n"
                    content.append(reactions_md)
                    logger.debug("Appended reactions section to markdown.")

            content.append("\n---\n")
            logger.debug("Added section separator to markdown.")

        # Write markdown file
        markdown_text = "\n".join(content)
        markdown_file_path = os.path.join(output_dir, filename_md)
        logger.info(f"Writing markdown file to: {markdown_file_path}")
        try:
            with open(markdown_file_path, "w", encoding="utf-8") as f:
                f.write(markdown_text)
            logger.info(f"Markdown file written successfully to: {markdown_file_path}")
        except Exception as e:
            logger.error(f"Error writing markdown file: {e}")

        logger.info(f"Thread exported successfully to {output_dir}")

    def _format_opsgenie_message(self, text):
        """
        Format Opsgenie message text for markdown export.

        :param text: Original message text
        :return: Formatted message text
        """
        # Replace Opsgenie-specific formatting with markdown formatting
        text = text.replace("h4.", "####")
        text = text.replace("* ", "* ")
        text = text.replace("[", "[").replace("|", "](").replace("]", ")")
        logger.debug(f"Formatted Opsgenie message: {text}")
        return text

    def _parse_slack_link(self, slack_link):
        """
        Parse Slack link to extract channel ID and thread timestamp.

        :param slack_link: Slack thread URL
        :return: Tuple of channel ID and thread timestamp
        """
        # Extract channel ID and thread timestamp using regex
        match = re.search(r"/archives/(\w+)/p(\d+)", slack_link)
        if not match:
            raise ValueError("Invalid Slack link format")

        channel_id = match.group(1)
        thread_ts = f"{match.group(2)[:10]}.{match.group(2)[10:]}"
        logger.debug(f"Extracted channel ID: {channel_id}, Thread TS: {thread_ts}")
        return channel_id, thread_ts

    def _sanitize_filename(self, filename):
        """
        Sanitize filename to remove any invalid characters.

        :param filename: Original filename
        :return: Sanitized filename
        """
        sanitized = "".join(
            [c for c in filename if c.isalpha() or c.isdigit() or c in (" ", ".", "_")]
        ).rstrip()
        logger.debug(f"Sanitized filename from '{filename}' to '{sanitized}'")
        return sanitized

    def _download_file(self, url, path, headers):
        """
        Download file from URL with authorization headers.

        :param url: File URL
        :param path: Path to save the file
        :param headers: Authorization headers
        :return: Boolean indicating success or failure
        """
        if not url:
            logger.error("No download URL provided.")
            return False

        try:
            logger.info(f"Attempting to download file from URL: {url}")
            response = requests.get(url, headers=headers, stream=True, timeout=10.0)
            logger.debug(f"Response Status Code: {response.status_code}")
            logger.debug(f"Response Headers: {response.headers}")

            if response.status_code == 200:
                content_type = response.headers.get("Content-Type", "").lower()
                # logger.debug(f"Content-Type of response: {content_type}")

                if "text/html" in content_type:
                    logger.error(
                        f"Unexpected Content-Type: {content_type}. Possible authentication issue."
                    )
                    return False

                with open(path, "wb") as f:
                    for chunk in response.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                logger.info(f"File downloaded successfully: {path}")
                return True
            else:
                logger.error(f"Failed to download file. Status Code: {response.status_code}")
                return False
        except Exception as e:
            logger.error(f"Exception occurred while downloading file: {e}")
            return False



import dataclasses
import datetime as dt
import logging
import typing as t
import urllib.parse
from abc import abstractmethod
from enum import Enum
from pathlib import Path

from attrs import define
from dataclasses_json import CatchAll, Undefined, dataclass_json

from rapporto.util import sanitize_title

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class ActivityInquiry:
    organization: t.Optional[str] = None
    created: t.Optional[str] = None
    author: t.Optional[str] = None


class QType(Enum):
    API = 1
    HTML = 2


class QKind(Enum):
    ISSUE = 1
    PULLREQUEST = 2


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclasses.dataclass
class PullRequestMetadata:
    # Original fields from GitHub API.
    number: str
    url: str
    html_url: str
    title: str
    commits: int
    additions: int
    deletions: int
    changed_files: int
    comments: int
    review_comments: int

    # Computed fields.
    code_size: t.Optional[int] = None
    comments_total: t.Optional[int] = None
    repo_name: t.Optional[str] = None

    # This dictionary includes all the remaining fields.
    more: t.Optional[CatchAll] = None

    def __post_init__(self):
        self.code_size = self.additions - self.deletions
        self.comments_total = self.comments + self.review_comments
        try:
            self.repo_name = self.more["base"]["repo"]["name"]  # type: ignore[index]
        except KeyError as ex:
            msg = f"Unable to decode repository name: {ex}"
            logger.error(msg)
            raise KeyError(msg) from ex

    def format_item(self):
        # Generic info (for debugging)
        # return (f"files: {self.changed_files}, "
        #        f"size: {self.code_size}, comments: {self.comments_total}")
        # Repo: PR+link
        return f"  - {self.repo_name}: [{sanitize_title(self.title)}]({self.html_url})"


class GitHubQueryBuilder:
    template_api = (
        "https://api.github.com/search/issues?q={query}&per_page=100&sort=created&order=asc"
    )
    template_html = "https://github.com/search?q={query}&per_page=100&s=created&o=asc"

    def __init__(self, inquiry: ActivityInquiry):
        self.inquiry = inquiry
        self.type: t.Optional[QType] = None
        self.kind: t.Optional[QKind] = None
        self.constraints: t.List[str] = []
        self.query()

    def add(self, field: str, value: t.Optional[str] = None):
        if value is not None:
            self.constraints.append(f"{field}:{value}")

    @property
    def expression(self):
        return " ".join(self.constraints)

    @abstractmethod
    def query(self):
        raise NotImplementedError("Needs to be implemented")

    @property
    def timerange(self):
        timerange = self.inquiry.created
        if timerange and "W" in timerange:
            p_year, p_week = timerange.split("W")
            timerange = self.date_range_str(*self.date_range_from_week(p_year, p_week))
        return timerange

    @staticmethod
    def date_range_from_week(p_year, p_week):
        """
        http://mvsourcecode.com/python-how-to-get-date-range-from-week-number-mvsourcecode/
        """
        firstdayofweek = dt.datetime.strptime(f"{p_year}-W{int(p_week) - 1}-1", "%Y-W%W-%w").date()
        lastdayofweek = firstdayofweek + dt.timedelta(days=6.9)
        return firstdayofweek, lastdayofweek

    @staticmethod
    def date_range_str(start: dt.datetime, end: dt.datetime):
        return f"{start.isoformat()}..{end.isoformat()}"

    @property
    def query_issues(self):
        return f"{self.expression} is:issue"

    @property
    def query_pulls(self):
        return f"{self.expression} is:pr"

    def issue(self):
        self.kind = QKind.ISSUE
        return self

    def pr(self):
        self.kind = QKind.PULLREQUEST
        return self

    def api(self):
        self.type = QType.API
        return self

    def html(self):
        self.type = QType.HTML
        return self

    def url(self):
        if self.type == QType.API:
            template = self.template_api
        elif self.type == QType.HTML:
            template = self.template_html
        else:
            raise NotImplementedError("Unknown type: Only API and HTML are supported")
        if self.kind == QKind.ISSUE:
            query = self.query_issues
        elif self.kind == QKind.PULLREQUEST:
            query = self.query_pulls
        else:
            raise NotImplementedError("Unknown kind: Only ISSUE and PULLREQUEST are supported")
        return template.format(query=urllib.parse.quote(query))


class GitHubActivityQueryBuilder(GitHubQueryBuilder):
    """
    Find all open issues and pull requests created by individual author.
    """

    def query(self):
        self.add("org", self.inquiry.organization)
        self.add("created", self.timerange)
        self.add("author", self.inquiry.author)


class GitHubAttentionQueryBuilder(GitHubQueryBuilder):
    """
    Find all open issues and pull requests with labels "bug" or "important".
    """

    labels = [
        "bug",  # GitHub standard.
        "important",  # CrateDB.
        "type-bug",  # CPython
        "type-crash",  # CPython
    ]

    def query(self):
        self.add("org", self.inquiry.organization)
        self.add("created", self.timerange)
        self.add("label", ",".join(self.labels))
        self.add("state", "open")


@define
class GitHubSearch:
    session: t.Any
    query_builder: GitHubQueryBuilder
    issues_api: str
    pulls_api: str
    issues_html: str
    pulls_html: str

    @classmethod
    def with_query_builder(cls, session, query_builder: GitHubQueryBuilder):
        return cls(
            session=session,
            query_builder=query_builder,
            issues_api=query_builder.issue().api().url(),
            pulls_api=query_builder.pr().api().url(),
            issues_html=query_builder.issue().html().url(),
            pulls_html=query_builder.pr().html().url(),
        )

    def issues_and_prs(self):
        items = []

        response = self.session.get(self.issues_api)
        response.raise_for_status()
        items += response.json()["items"]

        response = self.session.get(self.pulls_api)
        response.raise_for_status()
        items += response.json()["items"]

        return items


@dataclasses.dataclass
class MultiRepositoryInquiry:
    repositories: t.List[str]

    @classmethod
    def make(cls, repository: str, repositories_file: Path = None) -> "MultiRepositoryInquiry":
        if repository:
            return cls(repositories=[repository])
        elif repositories_file:
            return cls(repositories=repositories_file.read_text().splitlines())
        else:
            raise ValueError("No repository specified")


@dataclasses.dataclass
class ActionsFilter:
    event: str
    status: str
    created: str

    @property
    def query(self) -> str:
        return f"event={self.event}&status={self.status}&created={self.created}"


@dataclasses.dataclass
class ActionsOutcome:
    status: str
    repository: str
    name: str
    url: str
    started: str

    @property
    def markdown(self):
        title = sanitize_title(f"{self.repository}: {self.name}")
        return f"- [{title}]({self.url})"



try:
    from importlib.metadata import PackageNotFoundError, version
except (ImportError, ModuleNotFoundError):  # pragma:nocover
    from importlib_metadata import (  # type: ignore[assignment,no-redef,unused-ignore]
        PackageNotFoundError,
        version,
    )

try:
    __version__ = version("rapporto")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



import click

from rapporto.slack.core import SlackThreadExporter


@click.group()
@click.pass_context
def cli(ctx: click.Context):
    """
    Harvest information from Slack.
    """
    pass


@cli.command()
@click.argument("url", type=str, required=True)
@click.option(
    "--slack-token", type=str, envvar="SLACK_TOKEN", required=False, help="Slack API token"
)
def export(url: str, slack_token: str):
    """
    Export a Slack thread into Markdown format.
    """
    if not slack_token:
        raise click.UsageError(
            "Missing option '--slack-token' or 'SLACK_TOKEN' environment variable."
        )

    exporter = SlackThreadExporter(slack_token)
    exporter.export_thread(url)



import click
from click_aliases import ClickAliasedGroup
from pueblo.util.cli import boot_click

from rapporto.github.cli import cli as github_cli
from rapporto.slack.cli import cli as slack_cli


@click.group(cls=ClickAliasedGroup)
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    return boot_click(ctx, verbose, debug)


cli.add_command(github_cli, "gh")
cli.add_command(slack_cli, "slack")
